from .adept import download_adept, prepare_adept
from .aishell import download_aishell, prepare_aishell
from .aishell4 import download_aishell4, prepare_aishell4
from .ali_meeting import download_ali_meeting, prepare_ali_meeting
from .ami import download_ami, prepare_ami
from .aspire import prepare_aspire
from .babel import prepare_single_babel_language
from .broadcast_news import prepare_broadcast_news
from .bvcc import download_bvcc, prepare_bvcc
from .callhome_egyptian import prepare_callhome_egyptian
from .callhome_english import prepare_callhome_english
from .cmu_arctic import download_cmu_arctic, prepare_cmu_arctic
from .cmu_indic import download_cmu_indic, prepare_cmu_indic
from .cmu_kids import prepare_cmu_kids
from .commonvoice import prepare_commonvoice
from .cslu_kids import prepare_cslu_kids
from .dihard3 import prepare_dihard3
from .eval2000 import prepare_eval2000
from .fisher_english import prepare_fisher_english
from .fisher_spanish import prepare_fisher_spanish
from .gale_arabic import prepare_gale_arabic
from .gale_mandarin import prepare_gale_mandarin
from .gigaspeech import prepare_gigaspeech
from .heroico import download_heroico, prepare_heroico
from .hifitts import download_hifitts, prepare_hifitts
from .icsi import download_icsi, prepare_icsi
from .l2_arctic import prepare_l2_arctic
from .libricss import download_libricss, prepare_libricss
from .librimix import download_librimix, prepare_librimix
from .librispeech import download_librispeech, prepare_librispeech
from .libritts import download_libritts, prepare_libritts
from .ljspeech import download_ljspeech, prepare_ljspeech
from .mgb2 import prepare_mgb2
from .mls import prepare_mls
from .mobvoihotwords import download_mobvoihotwords, prepare_mobvoihotwords
from .mtedx import download_mtedx, prepare_mtedx
from .musan import download_musan, prepare_musan
from .nsc import prepare_nsc
from .peoples_speech import prepare_peoples_speech
from .rir_noise import download_rir_noise, prepare_rir_noise
from .spgispeech import download_spgispeech, prepare_spgispeech
from .switchboard import prepare_switchboard
from .tedlium import download_tedlium, prepare_tedlium
from .timit import download_timit, prepare_timit
from .vctk import download_vctk, prepare_vctk
from .voxceleb import download_voxceleb1, download_voxceleb2, prepare_voxceleb
from .wenet_speech import prepare_wenet_speech
from .yesno import download_yesno, prepare_yesno
